; Spc2Midi - DSP Emulation
;
; Note this only handles the emulation aspect of the DSP registers, not
; synthesizing output. That can be found in dspsim.asm. This mainly handles
; reacting to register writes and recording them in the buffer.
;
; todo: find what the real SPC does for samples with range >12 (StarFox!)
;       A lot SquareSoft games seem to have a time limitation to their songs.
;        Since they all share basically the same SPC core, and since I've
;        determined it isn't because the buffer is full, the core must be,
;        for some reason, unable to play songs for more than 28 minutes.
;        Need determine when this condition occurs.
;       don't record pitch waver used with most flute sounds (Zelda,FF3...)
;       do however record significant pitch slides (DKC Intro voice 3,
;        ChronoTrigger Frog's theme voice 1..). at least song (Vortex:
;        Voltair Voice2) uses pitch sliding exclusively to change notes.
;       determine if pitch is sliding or merely being set for the next kon
;        (Macross final boss voice 5, Bof1 battle most voices). some games
;        depend on the envelope decay to silence a voice (rather than kof)
;        before its pitch is set for the kon.
;       add some form of interpolation
;       figure out why dkcfunky.spc sample #34 sounds so bad playrate
;        3164 22khz (b20 Ad5 A60 937). maybe interpolation would fix it.
;       Vortext "Big Beast" voice 4 (vortex13.spc) pitch is zero!? perhaps
;        used as a convenient way to keep a voice off? SpcTool also shows it
;        as zero.
;       rudora~1.spc smp#14 clicks badly when paused/unpaused
;
; done: discovered clicking in some voices had nothing do with DSP emulation
;        after all, but a peculiarity of the Crystal Audio sound card.
;       figured out why VoiceEmu stalls half a second when rewinding
;       discovered why so many voices were silent. games have invalid ranges
;        or very short samples.
;       synced emulator timing to real SNES timing near exactly by putting
;        the two side by side
;       verified correct pitch by listening to the real thing some
;       average left & volume instead of only using left volume, so voices
;        that only use right volume will not be silent (terra.sp3 voice 5)
;
; DspEmu - handles register writes and recording to buffer
; DecompressAllSamples
; DecompressSample
; DecompressAdpcm
; GetAllSamplesInfo
; GetBrrSampleInfo
; GetCheckSum
;
;=============================================================================

DspReg:
; DSP special registers, all are both read and write
; (the following apply to each of the 8 voices)
.VolLeft                equ 0   ;       VOLL volume of left channel, 8bits signed
.VolRight               equ 1   ;       VOLR volume of right channel, 8bits signed
.PitchLow               equ 2   ;       PL low byte of 14bit sample pitch
.PitchHigh              equ 3   ;       PH high byte, 1000h is 32khz
.SourceNumber           equ 4   ;       SRCN sample to play 0-255
.AttackDecay            equ 5   ;       Attack and decay rate
.Sustain                equ 6   ;       Sustain level and rate
.Gain                   equ 7   ;       GAIN ??
  .Gain_Amount          equ 224 ;       bits 0-4
  .Gain_Type            equ 31  ;       bits 5-7
  .Gain_Direct          equ 0
  .Gain_IncLinear       equ 128+64
  .Gain_IncBentLine     equ 128+64+32
  .Gain_DecLinear       equ 128+0
  .Gain_DecExponent     equ 128+32
.Envelope               equ 8   ; r     ENVX
.Outx                   equ 9   ; r     ??
.Filter                 equ 0Fh ;
; (these apply to all channels)
.MainVolLeft            equ 0Ch ;       MVOL L
.MainVolRight           equ 1Ch ;       MVOL R
.EchoVolLeft            equ 2Ch ;       EVOL L
.EchoVolRight           equ 3Ch ;       EVOL R
.KeyOn                  equ 4Ch ;       KON voices 0-7, turn on
.KeyOff                 equ 5Ch ;       KOF voices 0-7, mute
.Flags                  equ 6Ch ;       FLG
  .Flags_SoftReset      equ 128
  .Flags_MuteAll        equ 64  ;       turns off all voices when true
  .Flags_Ecen           equ 32  ;       ECEN external memory echo ??
  .Flags_NoiseClock     equ 31  ;       NCK bits 0-4, noise generator clock
.Endx                   equ 7Ch ;       ENDX voices 0-7, flag voice sample has ended
.EchoFeedBack           equ 0Dh ;       EFB voices 0-7
;.Unused                equ 1Dh ;       --
.PitchModulation        equ 2Dh ;       PMOD voices 1-7 (voice 0 not affected)
.NoiseOnOff             equ 3Dh ;       NON voices 0-7
.EchoOnOff              equ 4Dh ;       EON voices 0-7
.SourceDirectory        equ 5Dh ;       DIR address of source directory * 256
.EchoDirectory          equ 6Dh ;       ESA address of echo region * 256 ??
.EchoDelay              equ 7Dh ;       EDL bits 0-3 ??


;
section .text

DspEmu:
;
; What each register holds during emulation:
;
; al  = emulated flags
; ah  = temp flags, otherwise free
; bl  = A
; bh  = Y
; bx+ = instruction counter    (upper 16 bits)
; ecx = free for opcodes
; ch  = used byte read/write to return and set bytes
; dx  = free for opcode pointer use   (upper 16 bits must remain zero!)
; si  = current opcode pointer (upper 16 bits must remain zero!)
; edi = free for opcodes and register functions
; ebp = keeps current time as cycle counter
;
; X, direct page, and stack pointer are not kept in registers.
;
; Opcode routines may use ecx,edx,edi,ah.
; Read/write routines may only use ah,edi. All others must be saved.
; edx must be reset to F3h before returning (address of the dsp data
; register) before returning.
; Some instructions require the use of specific registers, such as cl for
; shifting and ax for multiplying. For these, the registers may be saved
; (push) or temporarily swapped with another (xchg).

;
.Init:
    ; Expands the rle compressed pitch table, from 128 words to ~22k
    ; Each hertz frequency is rounded to the nearest MIDI note from 0 to 127.
    cld
    mov esi,HertzNoteRle
    xor ecx,ecx             ;zero extend counter
    mov edi,HertzNoteTbl
    xor eax,eax             ;zero current MIDI note
.NextNote:
    mov cx,[esi]            ;get RLE length
    rep stosb               ;fill table with current note
    add esi,byte 2          ;next strip
    inc al                  ;next MIDI note
    jns .NextNote           ;stop on 256 wrap

; Call this when loading a new savestate or SPC.
.Reinit:
    ; Clear all DSP sample attributes
    ; have all indexes point to "Default"
    cld
    xor eax,eax

    mov edi,SampleList
    mov ecx,SampleList.ClearSize/4
    rep stosd

    mov edi,BrrList
    mov ecx,BrrList.ClearSize/4
    rep stosd

    ;Clear BRR sample cache
    ;all samples as being unused
    mov edi,SamplesBuffer
    mov ecx,SamplesBufferSize/4
    rep stosd

    mov edi,VoiceEmu.Vars
    mov ecx,VoiceEmu.VarsSize/4
    rep stosd

    mov edi,FmEmu.Vars
    mov ecx,FmEmu.VarsSize/4
    rep stosd

    mov edi,MpuEmu.Vars
    mov ecx,MpuEmu.VarsSize/4
    rep stosd

    mov edi,SineEmu.Vars
    mov ecx,SineEmu.VarsSize/4
    rep stosd

    bts dword [SampleList.Used],eax
    mov dword [SampleList.Length],DefaultSample.Length
    mov dword [SampleList.LoopLen],DefaultSample.LoopLen
    mov dword [SampleList.Checksum],DefaultSample.CheckSum
    mov dword [SampleList.BrrIdx],DefaultSample.BrrIdx
    mov dword [SampleList.Instrument],DefaultSample.Instrument
    mov dword [SampleList.SoundFx],DefaultSample.SoundFx
    mov dword [SampleList.Freq],DefaultSample.Freq
    mov byte  [SampleList.Volume],DefaultSample.Volume
    mov byte  [SampleList.Interpolation],DefaultSample.Interpolation
    mov dword [SampleList.Name],eax

    ret

;    dd SampleList
;    dd SampleList.ClearSize/4
;    dd BrrList
;    dd BrrList.ClearSize/4
;    dd SamplesBuffer
;    dd SamplesBufferSize/4
;    dd VoiceEmu.Vars
;    dd VoiceEmu.VarsSize/4
;    dd 0


;
align 4
.ReadRegJumpTable:
; (empty for now)

.WriteRegJumpTable:
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg4C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg5C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg7C,.wReg0D,.wReg0E,.wReg0F
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX

;
; (edx=register read/written, ch=value read/written)
; (edx=F3h; !ah)
.wReg03: ;pitch, high 6bits of 14bit pitch, 1000h is 32khz
    and ch,3Fh
.wReg02: ;pitch, low 8bits of 14bit pitch, 1000h is 32khz
%if SupportPitchSlide
    mov edi,edx
    cmp [DspRam+edx],ch
    je .wReg02Same
    shr edi,4                   ;get voice #
    push ebx
    mov bh,[DspRam+edx]         ;save new value to register
    bt dword [DspRam+DspReg.KeyOn],edi
    mov [DspRam+edx],ch         ;save new value to register
    jnc .wReg02Ignore           ;ignore pitch change if voice off
    mov [DspRamCopy+edx],ch
    mov bl,ch
    ;mov edi,[DspBufferEnd]     ;get position in dsp buffer
    push edx
    call .RecordDspTime
    pop edx
    call .RecordDspRegVal
    mov [DspBufferEnd],edi      ;save position in dsp buffer
.wReg02Ignore:
    pop ebx
.wReg02Same:
    mov edx,0F3h                ;restore DSP data address
    ret
%endif

.wReg00:;left volume
.wReg01:;left volume
.wReg04:;sample source number
.wReg05:
.wReg06:
.wReg07:
.wReg08:
.wReg09:
.wReg0A:
.wReg0B:
.wReg0C:
.wReg0D:
.wReg0E:
.wReg0F:
    mov [DspRam+edx],ch         ;save new value to register
.wReg7C:                        ;ignore endx for now
.wIgnore:
    mov edx,0F3h                ;restore DSP data address
    ret

%if 0
.wReg7C:;endx
    mov [DspRam+DspReg.Endx],byte 0  ;clear endx bits
    mov edx,0F3h                ;restore DSP data address
    ret
%endif

;
.wReg4C:;key on
    test ch,ch                  ;a lot of games send pointless commands
    jz .wIgnore                 ;like turning on zero keys
    push ebx

    call .RecordDspTime
;    cmp edx,64000               ;ignore first second of play
;    mov ebx,edx
;    jb .MinSame
;    sub ebx,[DspBufferKonTime]
;    cmp [DspBufferMinRes],ebx
;    jbe .MinSame
;    mov [DspBufferMinRes],ebx
;.MinSame:
;    mov [DspBufferKonTime],edx

    ; turn off any keys already playing before turning them on again
    mov dh,ch
    and dh,[DspRam+DspReg.KeyOn]
    jz .KonAllNew
    ;xor [DspRam+DspReg.KeyOn],dh (unnecessary)
    mov dl,DspReg.KeyOff        ;restore key_off address
    call .RecordDspAndRet
.KonAllNew:

    ; set kon & kof
    mov ah,ch
    or [DspRam+DspReg.KeyOn],ch     ;set new keys on
    not ah
    and [DspRam+DspReg.KeyOff],ah   ;clear corresponding keys off

    ; Record pitch, volume, and brr source # in DSP buffer
    xor edx,edx                 ;start with volume register of first voice
    not ah                      ;which voices to turn on
    jmp short .KonCheck
.KonNext:
    call .RecordDspReg          ;left volume
    inc edx
    call .RecordDspReg          ;right volume
    inc edx
    call .RecordDspReg          ;pitch low
    inc edx
    call .RecordDspReg          ;pitch high
    inc edx
    call .RecordDspReg          ;source number
    movzx ebx,byte [DspRam+edx] ;get channel sample number
    add edx,byte 12             ;next voice
    bts [BrrList.Used],ebx      ;mark sample as being used
    jc .KonCheck
    bts [BrrList.Valid],ebx     ;if used by the game, then it MUST be valid (even if it breaks of the rules)
    bts dword [DspFlags],DspFlags.NewValidVoiceb
.KonCheck:
    shr ah,1                    ;was key turned on?
    jc .KonNext
    lea edx,[edx+16]
    jnz .KonCheck               ;any more keys turned on?

    pop ebx

    ; Record key on bits
    mov dl,DspReg.KeyOn         ;restore key_on address
    mov dh,ch                   ;copy value sent to register
    jmp short .RecordDspAndRet

%if 0
    call .RecordDspAndRet   ;was trying to figure out a voice that used
    pusha                   ;pitch sliding in Vortex
  .wait:
    mov ah,1
    int 16h
    jz .wait
    popa
    ret
%endif

;
.wReg5C:;key off
    mov ah,[DspRam+DspReg.KeyOn];compare channels currently on
    and ah,ch                   ;with channels to turn off
    jz near .wIgnore            ;skip if command affects nothing
    xor [DspRam+DspReg.KeyOn],ah;clear any channels turned off
    or [DspRam+DspReg.KeyOff],ah;set voices as off

    call .RecordDspTime

    mov dl,DspReg.KeyOff        ;restore key_off address
    mov dh,ah                   ;copy value sent to register
    ;jmp short .RecordDspAndRet (fall through)

;
; Records the value being written to a DSP register. Can either be called
; or jumped to. This routine does not record the previous value of the
; register, only the new one.
;
; (dl=dsp register, dh=register value, ch=value written, edi=buffer ptr)
; (edx=F3h, edi+=4; *)
.RecordDspAndRet:
  %if SpcDebugIoPrint
    pusha
    mov edi,Text.DspWrite
    call SpcEmu.PrintBreakWithValue
    popa
  %endif

    cmp edi,DspBufferSize-4     ;do not write past end of buffer
    jb .RdtNotFull
    push dword .RdtEnd
    jmp short .BufferEnd
.RdtNotFull:

    or edx,0FF000000h
    mov [DspBuffer+edi],edx     ;record register address & value
    add edi,byte 4              ;next event slot
.RdtEnd:
    mov edx,0F3h                ;restore DSP data address
    mov [DspBufferEnd],edi      ;save position in dsp buffer
    ret

;
; Jumped to when buffer is full to insert an ending key off.
;
; (flags="cmp edi,DspBufferSize-4")
; (; edx)
.BufferEnd:
    ja .BufferFull
    push edx
    mov edi,[DspBufferTime]     ;get tick count of previous write
    mov edx,0FF000000h|DspReg.KeyOff
    inc edi
    mov dh,[DspRam+DspReg.KeyOn]
    mov [DspBufferTime],edi
    mov [DspBuffer+DspBufferSize-4],edx
    mov edi,DspBufferSize
    pop edx
.BufferFull:
    ret

;
; Records the tick count of write.
;
; ()
; (edi=buffer ptr+4, edx=dsp buffer time; *)
.RecordDspTime:
    mov edi,[DspBufferEnd]      ;get end position in DSP buffer
    mov edx,[SpcEmu.TotalTime]  ;get tick count of write
    cmp edi,DspBufferSize-4     ;do not write past end of buffer
    jae .BufferEnd
    cmp [DspBufferTime],edx     ;very, very unlikely
    je .RdtmEnd                 ;but just in case..
    mov [DspBufferTime],edx
    mov [DspBuffer+edi],edx     ;store tick time of dsp write
    add edi,byte 4
.RdtmEnd:
    ret


;
; Records the value from a specified DSP register, other than the one
; currently being written to. Only new DSP values are recorded. If the
; register value between now and the last call are the same, nothing happens.
;
; (edx=dsp register to get value from, edi=buffer ptr)
; (edi+=4; edx,!ebx)
.RecordDspReg:
    mov bl,[DspRam+edx]
    mov bh,[DspRamCopy+edx]
    cmp bl,bh
    je .RdrEnd
    mov [DspRamCopy+edx],bl
; (bl=current value, bh=previous value)
.RecordDspRegVal:               ;call here for no check
  %if SpcDebugIoPrint
    push ecx
    mov ch,[DspRam+edx]
    pusha
    mov edi,Text.DspWrite
    call SpcEmu.PrintBreakWithValue
    popa
    pop ecx
  %endif
    cmp edi,DspBufferSize-4     ;do not write past end of buffer
    jae .BufferEnd

    shl ebx,8                   ;move previous register value to third byte
    or ebx,edx                  ;copy register address to low byte
    or ebx,0FF000000h           ;flag that dword is a register write (and not timer value)
    mov [DspBuffer+edi],ebx     ;save both register address and value
    add edi,byte 4              ;next entry
.RdrEnd:
    ret

;
.wRegXX:
    pusha
%if SpcDebugMode != 1
	mov [SpcEmu.AY],bx
	mov [SpcEmu.Pc],si
	mov [SpcEmu.Flags],al
    mov edi,MakeOperandString.Buffer
	call MakeOperandString
    mov edx,MakeOperandString.Buffer
	call WriteString
%endif
    mov edi,Text.DspWriteIllegal
    call SpcEmu.PrintBreakWithValue
    popa
    ret


;
; Reads all samples pointed to by the directory, setting their attributes.
; Also tests for validity use
;
; () (!)
GetAllSamplesInfo:
    mov esi,Text.ValidatingSamples
    call StatusMessage

    xor ebx,ebx
    mov bh,[DspRam+DspReg.SourceDirectory]  ;*256
    xor ecx,ecx                 ;set starting sample
    add ebx,SpcRam

.Next:
    ; validate sample before expanding into wave buffer
    push ecx                    ;save sample number
    mov esi,[ebx]               ;get sample offset from table
    push ebx                    ;save source directory ptr
    push esi                    ;save sample offsets

    and esi,65535               ;bottom word only for sample
    call GetBrrSampleInfo
    mov [.Loop],al              ;save header byte
    lea edi,[esi+ecx]           ;figure end byte
    push edx                    ;save sample length
    add esi,SpcRam              ;GetCheckSum requires an absolute address
    call GetCheckSum
    test byte [.Loop],2         ;check header byte for if sample is looped
    jnz .Looped

    cmp [esp],dword 16*3        ;sample must be at least 3 blocks long (48 samples)
    jae .SizeOk
    inc ebx                     ;or else consider it garbage
.SizeOk:
    push dword 0                ;zero loop length
    jmp short .SampleEnd

.Looped:
    movzx esi,word [esp+6]      ;retrieve sample offset of loop portion

    ; check that main and loop portions are aligned
    sub edi,esi                 ;last byte after main <= loop portion
    jbe .LoopSeparate
    cmp [esp+4],si              ;starting portion > loop portion
    ja .LoopSeparate
    ; (eax=checksum, edi=loop length in bytes)
    xchg eax,edi                ;swap checksum with byte after main portion
    mov ecx,9                   ;9 bytes per 4bit ADPCM block
    xor edx,edx
    div ecx
    test edx,edx
    mov eax,edi                 ;restore checksum into eax again
    jz .LoopSeparate
    inc ebx                     ;main portion and loop portion aren't aligned

.LoopSeparate:
    push eax                    ;save checksum of main portion
    call GetBrrSampleInfo.StartingWith
    mov [.Loop],al              ;save header byte
    pop eax                     ;get previous checksum of sample
    push edx                    ;save sample length of loop portion
    add esi,SpcRam              ;GetCheckSum requires an absolute address
    call GetCheckSum.StartingWith
    test byte [.Loop],2         ;check header byte for if sample is looped
    jnz .SampleEnd              ;if it is, then ok
    inc ebx                     ;else loop portion is not looped!
.SampleEnd:

    ; if no errors in either main or loop portion, mark as valid
    mov edx,[esp+16]            ;get sample number
    test ebx,ebx
    jnz .SampleInvalid
    bts [BrrList.Valid],edx
.SampleInvalid:

    shl edx,2
    mov [BrrList.Checksum+edx],eax      ;save checksum of both sample and loop portion
    pop dword [BrrList.LoopLen+edx]     ;set sample loop length in table
    pop dword [BrrList.Length+edx]      ;set sample length in table
    pop dword [BrrList.Offset+edx]      ;set offsets of sample and loop portion

    pop ebx
    pop ecx
    add ebx,byte 4
    inc cl
    jnz near .Next

    ret

section .bss
.Loop:  resb 1
section .text


;
; Buffers all 256 samples from source directory into sound wave buffer,
; checking that each one is valid before expanding.
;
; () (!)
DecompressAllSamples:
    mov esi,Text.BufferingSamples
    call StatusMessage

; if valid sample
;   expand into wave buffer
; endif
    xor ecx,ecx
.Next:
    bt [BrrList.Valid],ecx
    jnc .SkipSample
    push ecx
    call DecompressSample
    pop ecx
.SkipSample:
    inc cl
    jnz .Next

    ret

;
; Decompresses a whole sample, including both the main and loop portion of it.
; Also sets its attributes in the table. I've never seen it before, but I
; imagine some games could store the loop portion of a sample separately from
; the starting portion, perhaps so multiple samples could share the same
; sustain..?
;
; (dword sample number 0-255)
DecompressSample:
    xor ebx,ebx
    mov ecx,[esp+4]
    mov bh,[DspRam+DspReg.SourceDirectory]  ;*256
    mov esi,[SpcRam+ebx+ecx*4]  ;get sample address from directory table

    ; expand starting sample into wave buffer
    push esi                    ;save sample offsets
    and esi,65535               ;bottom word only for sample
    call DecompressAdpcm
    lea ebx,[esi+ecx]           ;figure end byte
    shr edi,1                   ;cache ptr /2
    pop esi                     ;retrieve sample offsets
    test al,2
    push edi                    ;save sample cache ptr
    jnz .Looped
    push dword 0                ;null cache ptr for loop portion
    jmp short .SampleEnd

.Looped:

    ; expand loop portion into wave buffer
    shr esi,16                  ;sample offset of loop portion is top word
    cmp ebx,esi                 ;last byte after main <= loop portion
    jbe .LoopSeparate
    cmp [esp+8],si              ;starting portion > loop portion
    ja .LoopSeparate

    ; loop portion is part of main sample (don't bother to decompress again)
    ; (eax=byte after end of main portion)
    mov eax,esi
    mov ecx,9                   ;9 bytes per 4bit ADPCM block
    xor edx,edx
    div ecx
    shl eax,4                   ;*16
    push eax
    jmp short .SampleEnd

.LoopSeparate:
    call DecompressAdpcm.Loop
    shr edi,1                   ;cache ptr /2
    push edi                    ;save sample loop cache ptr
.SampleEnd:

    mov ecx,[esp+12]            ;get sample number
    pop dword [BrrList.LoopPtr+ecx*4]   ;set loop portion cache ptr
    pop dword [BrrList.SamplePtr+ecx*4] ;set main sample portion cache ptr
    ret

.Loop   equ GetAllSamplesInfo.Loop


;
; Decompresses a 4bit ADPCM sample into the buffer. Filters are implemented.
;
; Returns the length of only the sample given. For the length of the loop
; portion too, the loop portion offset in RAM must be passed to it separately
; then the length of the original sample and loop portion can be added
; together to get the total sample length.
;
; This routine assumes the sample has already been checked for validity and
; isn't just random data (which makes a nice static sound). If it is invalid,
; the routine will happily decompress the garbage data anyway (just in case
; it really is valid wave data but was misidentified).
;
; What to do about blocks with invalid ranges with seems to be a matter of
; disagreement among various SPC players. I chose to clip the range to the
; maximum 12, but later found that SpcTool sets the range to zero, while
; ZSNES uses the valid shift of the previous block. Only the real SNES can
; say which is best.
;
; (esi=sample offset in SPC RAM)
; (eax=header byte of last block
;  ecx=byte length,
;  edx=sample length,
;  eax=last header byte,
;  esi=sample offset unchanged
;  edi=buffer offset)
DecompressAdpcm:
    ; zero previous samples used by filters
    xor ecx,ecx
    mov dword [.PrevSample],ecx ;previous sample before current one
    mov dword [.PrevSample2],ecx;sample two back from current one

.Loop:
    ; calculate desination in buffer (sample_offset / 9 * 32)
    push ebp
    mov eax,esi                 ;for the division
    mov ebx,9
    xor edx,edx
    push esi                    ;save sample base for later
    div ebx                     ;/9
    push dword 0                ;set sample length counter to zero
    shl eax,5                   ;*32
    push eax                    ;save cache offset ptr
    lea edi,[SamplesBuffer+eax]

    xor eax,eax
    mov ebx,[.PrevSample]       ;previous sample before current one
    align 16
.Next:
;(eax<=255!, esi=brr source, edi=buffer dest)
    mov al,[SpcRam+esi]         ;get header byte
    inc esi
    mov ecx,eax                 ;copy header byte
    and eax,15                  ;isolate filter and end bits only
    shr ecx,4                   ;get range from top four bits
    not ecx
    add ecx,29|(8<<8)           ;get 28-range  (!range+29)
    cmp cl,16
    jae .RangeOk
    mov cl,16
    ;mov cl,31
.RangeOk:
                                ;and set counter for eight bytes to follow
.NextByte:
    mov dl,byte [SpcRam+esi]    ;get high nybble
    and dl,0F0h                 ;mask out low nybble
    shl edx,24                  ;move high nybble to top of dword
    sar edx,cl                  ;apply range to high nybble

    ; if any filters needed
    ;   if two filters
    ;     filters = filter1 * [current sample-1]
    ;             - filter2 * [current sample-2]
    ;   else one filter
    ;     filters = filter1 * [current sample-1]
    ;   endif
    ;   sample += filters
    ; endif

    ; apply filter to high nybble
    cmp eax,1000b
    jb .NoSecondFilter1
    mov ebp,ebx                 ;copy previous sample
    xchg [.PrevSample2],ebx     ;get second sample back & swap with previous
    imul ebp,[.FilterTbl+eax*4]
    imul ebx,[.FilterTbl2+eax*4]
    add ebx,ebp                 ;filter1 + -filter2
    jmp short .AddFilter1
.NoSecondFilter1:
    ;mov ebx,[.PrevSample]      ;(unnecessary)
    mov [.PrevSample2],ebx
    cmp eax,100b
    jb .NoFilter1
    imul ebx,[.FilterTbl+eax*4]
.AddFilter1:
    sar ebx,8                   ;/256
    add edx,ebx                 ;sample value + filters
    cmp edx,32768               ;check for sample overflow
    jb .NoFilter1               ;-32768 - 32767
    cmp edx,-32768
    jae .NoFilter1
    sar edx,31                  ;extend sign down
    xor edx,32767
.NoFilter1:
    ;mov [.PrevSample],edx      ;(unnecessary)
    mov [edi],dx                ;store high nybble first

    mov bl,byte [SpcRam+esi]    ;get low nybble
    shl ebx,28                  ;move low nybble to top of dword
    sar ebx,cl                  ;apply range to low nybble

    ; apply filter to low nybble
    cmp eax,1000b
    jb .NoSecondFilter2
    mov ebp,edx                 ;copy previous sample
    xchg [.PrevSample2],edx     ;get second sample back & swap with previous
    imul ebp,[.FilterTbl+eax*4]
    imul edx,[.FilterTbl2+eax*4]
    add edx,ebp                 ;filter1 + -filter2
    jmp short .AddFilter2
.NoSecondFilter2:
    ;mov edx,[.PrevSample]      ;(unnecessary)
    mov [.PrevSample2],edx
    cmp eax,100b
    jb .NoFilter2
    imul edx,[.FilterTbl+eax*4]
.AddFilter2:
    sar edx,8                   ;/256
    add ebx,edx                 ;sample value + filters
    cmp ebx,32768               ;check for sample overflow
    jb .NoFilter2               ;-32768 - 32767
    cmp ebx,-32768
    jae .NoFilter2
    sar ebx,31                  ;extend sign down
    xor ebx,32767
.NoFilter2:
    ;mov [.PrevSample],ebx      ;(unnecessary)
    mov [edi+2],bx              ;store low nybble last

    inc esi
    add edi,byte 4
    dec ch
    jg near .NextByte

    add dword [esp+4],byte 16   ;sample count+=16
    test eax,1
    jnz .BrrEnd
    cmp esi,65536-9
    jbe near .Next
.BrrEnd:

    mov [.PrevSample],ebx       ;for next call (if loop exists)
    mov ecx,esi
    pop edi                     ;cache offset ptr
    pop edx                     ;retrieve length in samples
    pop esi                     ;retrieve sample base
    pop ebp
    sub ecx,esi                 ;[OR] byte_length = starting_offset - ending_offset
    ret


section .bss
alignb 4
.PrevSample:    resd 1          ;previous sample before current one
.PrevSample2:   resd 1          ;sample two back from current one
section .data
align 4
; All filter ratios are scaled up to a common denominator of 256
; 15/16=240/256  61/32=488/256  115/64=460  13/16=208/256
.FilterTbl equ $-16             ;don't store any filter values for filter 0
    dd 240,240,240,240, 488,488,488,488, 460,460,460,460
.FilterTbl2 equ $-32            ;don't store any filter values for filter 0&1
    dd -240,-240,-240,-240, -208,-208,-208,-208
section .text


;
; Returns info about a portion of BRR data including: byte length, its last
; header, and the data's validity.

; Performs several tests on the sample to determine whether it is valid or
; just random junk, returning the number of errors found.
;
; Disqualifying conditions include:
;   Starting in direct page RAM (0-1FFh)
;   Overlapping the ROM (FFC0h-FFFFh)
;   Invalid shift range (>12)
;   Loop bit inconsistently set
;   If part of main sample, loop portion must align
;   Unlooped sample to short (<48)
;   Loop portion does not have loop bit set
;
; Hopefully there are no games that violate any of these expected rules.
; In some ways, Spc2Midi is much more selective than SpcTool, which helps
; eliminate more garbage samples, but in other cases, SpcTool is more
; disriminating.
;
; Already found out that at least one game that breaks a rule. StarFox has a
; sample with a shift range of 13, and a loop portion of only 16 samples.
;
; Some of these rules are implemented by this routine; other rules are tested
; elsewhere.
;
; Other possible conditions unimplemented:
;   Loop portion not being within main sample
;   Using a filter in the first block
;
; Generally the loop portion is simply the tail end of the main sample, but
; there is no guarantee that all games do this. It is indeed possible that
; a game could share the same sustain for two sounds, but have different
; attacks. If, however, the loop is within main sample, the loop must align
; to it (evenly divisible by 9).
;
; Although it would seem wrong to use a filter in the very first block, it
; certainly is possible and shouldn't rule out a sample as valid if a game
; does. I simply zero the two 'previous' point values.
;
; Very rarely are samples shorter than 48 points. You can almost be certain
; (almost) that any unlooped sample only 16 or 32 points is invalid. As a
; matter of fact, the only sample I've ever that was so short was a looped
; 16 point square wave from Star Fox.
;
; (esi=sample offset in SPC RAM)
; (al=header byte of last block
;  ebx=number of errors,
;  ecx=sample byte length,
;  esi=sample offset unchanged)
%if 0
ValidateBrrSample:
    cmp esi,256                 ;sample should not start in direct page RAM
    jb .Invalid
    mov ah,[SpcRam+esi]         ;get header byte
    mov ecx,esi
    mov al,ah
    align 16
.CheckNext:
    xor al,ah
    test al,2                   ;is loop bit inconsistently set?
    jnz .Invalid
    cmp ecx,65463               ;sample should not overlap ROM
    ;cmp ecx,65536-9             ;sample should not wrap RAM
    ja .Invalid
    mov al,[SpcRam+ecx]         ;get header byte
    add ecx,byte 9              ;next BRR block
    cmp al,(13<<4)|15           ;range must be valid 0-12 (changed to 13 now)
    ja .Invalid                 ;some StarFox samples contain invalid ranges!
    test al,1
    jz .CheckNext
    sub ecx,esi
    ;(zero flag always clear)
    ret

.Invalid:
    xor ecx,ecx                 ;set byte length to 0
    ;(zero flag set)
    ret

%else

GetBrrSampleInfo:
    xor ebx,ebx                 ;initially no invalidating errors
.StartingWith:
    cmp esi,512                 ;sample should not start in direct page RAM
    jae .DpValid
    inc ebx
  .DpValid:
    mov ah,[SpcRam+esi]         ;get header byte
    xor edx,edx                 ;zero point length
    mov ecx,esi
    jmp short .LoopValid
    align 16
.CheckNext:
    xor al,ah
    test al,2                   ;loop bit consistently set?
    jz .LoopValid
    inc ebx
  .LoopValid:
    cmp ecx,65463               ;sample should not overlap ROM
    jbe .RomValid
    inc ebx
    jmp short .Abort
    ;cmp ecx,65536-9            ;sample must not wrap RAM
    ;ja .Abort
  .RomValid:
    mov al,[SpcRam+ecx]         ;get header byte
    add ecx,byte 9              ;next BRR block
    cmp al,(13<<4)|15           ;range must be valid 0-12 (changed to 13 now)
    jbe .ValidRange             ;some StarFox samples contain invalid ranges!
    inc ebx
  .ValidRange:
    add edx,byte 16             ;count another 16 sample points from block
    test al,1
    jz .CheckNext

.Abort:
    sub ecx,esi
    ;test ebx,ebx               ;set zero flag
    ret
%endif


;
; This function's purpose is for uniquely identifying a given sample,
; although it could actually checksum anything. The checksum that it returns
; is the dword sum of the raw sample, not the expanded sample. The final
; dword (for any samples with lengths not an even multiple of four) will be
; masked to exclude the last few bytes which are not part of the sample, then
; added to the checksum. For a looped sample, this function must be called
; twice for a complete checksum, once for the sample beginning and again for
; the loop portion.
;
; (esi=source, ecx=byte length)
; (eax=addition checksum; ebx,edi)
GetCheckSum:
    xor eax,eax                 ;zero checksum
;(eax=existing checksum)
.StartingWith:
    mov edx,ecx                 ;copy length
    shr ecx,2                   ;/4 get dwords
    jz .LastBytes
.NextDword:
    add eax,[esi]               ;add to checksum
    add esi,byte 4
    loop .NextDword
.LastBytes:
    and edx,3                   ;%4 get leftover bytes
    jz .NoLastBytes
    shl edx,3                   ;*8
    bts ecx,edx                 ;set bit in ecx (which is zero from the loop above)
    dec ecx                     ;make mask, = 1 << ((sampleLength and 3) * 8)
    and ecx,[esi]               ;get final dword
    add eax,ecx                 ;add final few bytes to checksum
.NoLastBytes:
    ret
