;Spc2Midi - MIDI file output
;============================================================

;==============================
section .data
Midi:
.HeaderLen      equ 22  ;MIDI header (14) + Track header (8)
.Header:
    ;MIDI Header template
    ;Header ID, Length of six bytes, Format 0, Tracks 1, Delta time of 250
    db "MThd",0,0,0,6,0,0,0,1,0,250
    ;Followed by the track header, which the size of is later filled in,
    ;since the track size is unknown initially. After done recording, it is
    ;set by CloseMidi.
    db "MTrk",0,0,0,0
.TempoLen       equ 7
.Tempo:
    ;Tempo of one beat every 500,000 microseconds (two beats per second)
    db 0,0FFh,51h,3,7,0A1h,20h
.EndOfTrackLen  equ 4
.EndOfTrack:
    db 0,0FFh,02Fh,0

;==============================
section .bss
.FileBufferSize     equ 2048
.Instruments:       resb 16 ;holds the last used instrument for that channel
.NoteBufferIndex:   resd 1  ;point read up to in NoteBuffer
.FileNamePtr:       resd 1
.FileHandle:        resd 1
.FileSize:          resd 1  ;file size in bytes, not including headers
.FileBufferIndex:   resd 1  ;point to write to next (also the number of bytes in it)
.FileBuffer:        resb Midi.FileBufferSize+32


;============================================================
section .text

.NoteOff        equ 080h
.NoteOn         equ 090h
.KeyAfterTouch  equ 0A0h
.ControlChange  equ 0B0h
.PatchChange    equ 0C0h
.ChanAfterTouch equ 0D0h
.PitchBend      equ 0E0h
.MetaEvent      equ 0F0h


;==============================
; Creates a MIDI file and writes the initial header.
; Also resets all variables in the Midi section.
; Note that this function will overwrite any existing file of the given name.
;
; (esi=filename) (cf=error) (saves none)
CreateMidi:
    ;open MIDI file
    mov ah,3Ch                      ;create file, only writing is necessary
    mov ecx,32                      ;set archive attribute
    mov edx,esi                     ;filename
    int 21h
    jc near .End                    ;error opening
    mov [Midi.FileHandle],eax

    ;A note on timing:
    ;The average precision for most games is 1/500th of a second, as they
    ;usually set timer 0 (8000khz) to count once every 16 ticks. There is no
    ;need to have the delta time ticks more precise than this, since it would
    ;be humanly undetectable anyway. The SPC700 (or least this emulated
    ;version of one) runs at 2496000 cycles per second, meaning that events
    ;would be rounded to the nearest 4992 cycles (2496000 / 500). The default
    ;MIDI beats per minute is 120, making 2 beats per second. So the delta
    ;time ticks per beat would be 500 ticks per second / 2 beats per second,
    ;or 250.

    ;ouput MIDI header
    mov ebx,eax                     ;copy file handle from the above open
    mov ecx,Midi.HeaderLen
    mov edx,Midi.Header
    mov ah,40h                      ;write to file
    int 21h
    jc .CloseEnd                    ;error writing header

    ;Reset variables
    xor eax,eax                     ;set to 0
    mov [Midi.NoteBufferIndex],eax
    mov [Midi.FileBufferIndex],eax
    mov [Midi.FileSize],eax
    dec eax                         ;set to -1
    cld
    mov edi,Midi.Instruments        ;-1 is an impossible instrument, so a
    mov ecx,4                       ;program change will always occur
    rep stosd                       ;the first time.
    mov [Midi.Instruments],eax      
    mov [Midi.Instruments+4],eax    
    mov [Midi.FileNamePtr],esi

    ;set default tempo of two beats per second
    ;120bpm is the default anyway, but I'm explicitly setting just in case
    mov esi,Midi.Tempo
    mov ecx,Midi.TempoLen
    call WriteMidiEvent

.End:
    ret

; (ebx=file handle)
.CloseEnd:
    mov ah,3Eh          ;close file
    int 21h
    stc    
    ret

;==============================
; Writes any unwritten MIDI events to file, turns off any notes still on,
; writes the end of MIDI track event, sets the size in the header, and
; closes the file.
;
CloseMidi:
    ;turn off any notes still on

    ;end MIDI with end of track event
    mov esi,Midi.EndOfTrack
    mov ecx,Midi.EndOfTrackLen
    call WriteMidiEvent
    call FlushMidiBuffer

    ;set track length
    mov eax,4200h       ;set file position
    mov ebx,[Midi.FileHandle]
    xor ecx,ecx
    mov edx,18          ;offset in file of track length
    int 21h
    mov ecx,[Midi.FileSize]
    xchg cl,ch          ;set to stupid big-endian order
    rol ecx,16
    xchg cl,ch
    push ecx
    mov ah,40h          ;write to file
    mov edx,esp         ;source is big-endian file size saved on stack
    mov ecx,4           ;dword
    int 21h
    pop ecx

    ;close MIDI file
    mov ah,3Eh          ;close file
    ;mov ebx,[Midi.FileHandle]
    int 21h
    ret

;==============================
; Reads the note buffer for new notes, converts them each into delta times and
; MIDI events, and records them to the file buffer.
; Since emulation is usually slightly ahead of actual time, or since it might
; be recording from the note buffer already filled,it only reads the note
; buffer up to notes that are less than or equal to the current playing time,
; not the cycles simply emulated.
;
OutputMidiEvents:
    ;start from last point in buffer
    ;read forward to end of buffer
    ;  if time that note should be played is less than cycle time then
    ;    if note is different from last note, set patch
    ;    if note is percussion, set channel to 10 and note value to instrument
    ;    commit the note to the MIDI file
    ;  elseif note time is greater than current time, exit loop

    ret

    mov edi,[Midi.FileBufferIndex]

    cmp edi,Midi.FileBufferSize
    jb .NoFlush
    mov [Midi.FileBufferIndex],edi
    call FlushMidiBuffer        ;start with new buffer
    xor edi,edi                 ;reset file buffer index
.NoFlush:

    ;writes big-endian delta time
    ;(eax=binary cycle time, edi=buffer offset)
    mov ecx,edi
    jmp short .FirstDeltaTime
.NextDeltaTime:
    mov bl,al
    or bl,80h
    shr eax,7
    shl ebx,8
    inc edi
.FirstDeltaTime:
    cmp eax,128
    jae .NextDeltaTime
    mov bl,al
    mov [Midi.FileBuffer+ecx],ebx

    ;get channel (voice number)
    ;if note on
    ;  mark channel as on
    ;  get sample number
    ;  get instrument
    ;  if instrument percussion (128-255)
    ;    set channel to 10
    ;    set note number to instrument-128
    ;  elif instrument different from last instrument for current channel
    ;    write patch
    ;    store status byte
    ;    write zero delta time
    ;  endif
    ;  if note number out of range, limit to 0-127
    ;  if run mode, check if status byte is the same
    ;  write Midi.NoteOn, note number, velocity
    ;  store status byte, note number
    mov ebx,Midi.PatchChange<<8
    or bl,dl
    mov bh,cl
    mov dword [Midi.FileBuffer+edi],ebx
    add edi,byte 3
    ;mov byte [Midi.FileBuffer+edi],0

    ;elif note off
    ;  mark channel as off
    ;  get status byte, note number
    ;  if run mode, if status byte is the same, velocity is zero
    ;  write Midi.NoteOff, note number, velocity
    ;endif

    mov [Midi.FileBufferIndex],edi
    ret

;==============================
; Can be called to write an individual event, such as the track title, author
; name, or end of MIDI. The delta-time must be included with the event.
; This is mainly called by outside routines. RecordMidiEvents directly handles
; small MIDI events that occur frequently, such as note on/off.
;
; (esi=source, ecx=event length)
WriteMidiEvent:
    ;if (FileBufferIndex + length of MIDI instruction > filebuffersize) then
    ;   write buffer to file
    ;endif
    ;copy event to buffer

    mov edi,[Midi.FileBufferIndex]
    lea eax,[edi+ecx]
    cmp eax,Midi.FileBufferSize
    jbe .NoFlush
    push ecx                    ;save length of event
    call FlushMidiBuffer        ;start with new buffer
    xor edi,edi
    pop ecx
.NoFlush:
    ;copy event to buffer
    add edi,Midi.FileBuffer
    ;cmp ecx,Midi.FileBufferSize
    ;jbe .EventSizeOk
    ;mov ecx,Midi.FileBufferSize
    ;.EventSizeOk
    cld
    rep movsb
    ret

;==============================
; Writes any events still in buffer to file and adjusts the midi filesize.
;
; () () (saves esi,edi)
FlushMidiBuffer:
    ;if BufferIndex > 0 then
    ;   write buffer to file for .FileBufferIndex length bytes
    ;   .FileSize += .FileBufferIndex
    ;   .FileBufferIndex = 0
    ;endif

    mov ecx,[Midi.FileBufferIndex]
    test ecx,ecx
    jz .End
    add [Midi.FileSize],ecx
    mov dword [Midi.FileBufferIndex],0
    mov ebx,[Midi.FileHandle]
    mov edx,Midi.FileBuffer
    mov ah,40h          ;write to file
    int 21h
.End:
    ret
